import sys
import time

from PIL import ImageGrab

sys.path.append("..")
from src import Mouse
from src import Screenshot
from src import Keybord
from src import QRCode
import pyscreeze
import pyautogui
import pyhutool

Mouse.click(500, 500)
# x = Mouse.size()
# x = Mouse.position()
# Mouse.leftClick(500, 500)

# Screenshot.screenshot('test.png')
# x = Screenshot.locateOnScreen('img_1.png')

# Keybord.keyDown('x')
# Keybord.keyUp('x')
# Keybord.hotkey('enter')

# QRCode.createQrcode('test', 'qrcode.png')



