import sys
import os
import time
import datetime
import qrcode
import hashlib

def createQrcode(content, file_path):
    qr = qrcode.QRCode(
        version=1,
        error_correction=qrcode.constants.ERROR_CORRECT_L,
        box_size=10,
        border=4,
    )
    qr.add_data(content)
    qr.make(fit=True)

    image = qr.make_image()
    image.save(file_path)